﻿using System;
using System.Text;
using System.IO;
using System.Diagnostics;
using System.Collections;
using Microsoft.SPOT;
using USBDbgLog;

namespace StepperMotorMidiPlayer
{
    public class MIDIFileParser
    {
        private byte[] fileBuffer;
        private uint index;

        public short TicksPerBeat { get; private set; }
        public int Tempo { get; private set; }
        public ArrayList NoteEvents { get; private set; }

        public struct NoteEvent
        {
            public uint DeltaTime;
            public byte EventType;
            public byte NoteNumber;
        }

        public MIDIFileParser()
        {
            
        }

        public void ParseFile(USBDrive drive)
        {
            this.NoteEvents = new ArrayList();
            this.Tempo = 500000;

            int trackNum = ReadFileIntoBuffer(drive);

            ReadHeaderChunk();

            for (int i = 0; i < trackNum; i++)
            {
                BurnTrackChunk();
            }

            ReadTrackChunk(this.index);
        }

        private int ReadFileIntoBuffer(USBDrive drive)
        {
            int trackNum = 0;

            fileBuffer = drive.ReadFile(ref trackNum);

            return trackNum;
        }

        private bool ReadHeaderChunk()
        {
            // Read "MThd"
            if (GetNextByte() == 0x4D &&
                GetNextByte() == 0x54 &&
                GetNextByte() == 0x68 &&
                GetNextByte() == 0x64)
            {
                // Read header length (should always be 6)
                if (GetNextByte() == 0x00 &&
                    GetNextByte() == 0x00 &&
                    GetNextByte() == 0x00 &&
                    GetNextByte() == 0x06)
                {
                    // Read format - 0 = single track, 1 = multiple track, 2 = multiple song
                    GetNextByte();
                    GetNextByte();

                    // Read number of tracks that follow the header chunk
                    GetNextByte();
                    GetNextByte();

                    // Read the division (ticks per beat)
                    this.TicksPerBeat = (short)((short)GetNextByte() << 8);
                    this.TicksPerBeat += GetNextByte();

                    return true;
                }
            }

            return false;
        }

        private void BurnTrackChunk()
        {
            uint trackLength;

            // Read "MTrk"
            if (GetNextByte() == 0x4d &&
                GetNextByte() == 0x54 &&
                GetNextByte() == 0x72 &&
                GetNextByte() == 0x6B)
            {
                // Read track length
                trackLength = (uint)((uint)GetNextByte() << 24);
                trackLength += (uint)((uint)GetNextByte() << 16);
                trackLength += (uint)((uint)GetNextByte() << 8);
                trackLength += GetNextByte();

                this.index += trackLength;
            }
        }

        private bool ReadTrackChunk(uint startIndex)
        {
            uint trackLength;

            // Read "MTrk"
            if (GetNextByte() == 0x4d &&
                GetNextByte() == 0x54 &&
                GetNextByte() == 0x72 &&
                GetNextByte() == 0x6B)
            {
                // Read track length
                trackLength  = (uint)((uint)GetNextByte() << 24);
                trackLength += (uint)((uint)GetNextByte() << 16);
                trackLength += (uint)((uint)GetNextByte() << 8);
                trackLength += GetNextByte();


                while (this.index < trackLength + startIndex)
                {
                    ReadTrackEvent();
                }

                return true;
            }

            return false;
        }

        private void ReadTrackEvent()
        {
            uint deltaTime = ReadDeltaTime();

            byte eventType = GetNextByte();

            if (eventType == 0xF0 ||
                eventType == 0xF7)
            {
                Debug.Print("Sys Ex Event");
                ReadSysExEvent();
            }
            else if (eventType == 0xFF)
            {
                Debug.Print("Meta Event");
                ReadMetaEvent();
            }
            else
            {
                Debug.Print("Midi Event");
                ReadMidiEvent(eventType, deltaTime);
            }
        }

        private byte runningStatus;
        private void ReadMidiEvent(byte eventTypeIn, uint deltaTime)
        {
            byte eventType;
            byte parm1, parm2;

            // Check for running status
            if ((eventTypeIn & 0x80) > 0)
            {
                eventType = eventTypeIn;
                runningStatus = eventType;
                parm1 = GetNextByte();
            }
            else
            {
                // running status, use last value
                eventType = runningStatus;
                parm1 = eventTypeIn;
            }

            

            NoteEvent noteEvent = new NoteEvent();
            noteEvent.DeltaTime = deltaTime;
            noteEvent.EventType = eventType;
            noteEvent.NoteNumber = parm1;

            if ((eventType & 0xF0) == 0x80)
            {
                // Note Off
                Debug.Print("Note Off: " + parm1.ToString());
                parm2 = GetNextByte();

                this.NoteEvents.Add(noteEvent);
            }
            else if ((eventType & 0xF0) == 0x90)
            {
                // Note On
                Debug.Print("Note On: " + parm1.ToString());
                parm2 = GetNextByte();

                this.NoteEvents.Add(noteEvent);
            }
            else
            {
                if (((eventType & 0xF0) == 0xC0) ||
                    ((eventType & 0xF0) == 0xD0))
                {

                }
                else
                {
                    parm2 = GetNextByte();
                }
                // Unhandled MIDI events
                Debug.Print("Unhandled MIDI Event: " + eventType.ToString());
            }
        }

        private void ReadMetaEvent()
        {
            byte metaType = GetNextByte();
            uint len = ReadVLength();

            if (metaType == 0x51)
            {
                this.Tempo = (int)((int)GetNextByte() << 16);
                this.Tempo += (int)((int)GetNextByte() << 8);
                this.Tempo += GetNextByte();
                Debug.Print("New Tempo: " + this.Tempo.ToString());
            }
            else
            {
                Debug.Print("Ignored Meta Event: Length=" + len.ToString());

                if (metaType == 0x2f)
                {
                    Debug.Print("End of Track");
                }

                this.index += len;
            }
        }

        private void ReadSysExEvent()
        {
            uint len = ReadVLength();

            for (int i = 0; i < len; i++)
            {
                GetNextByte();
            }
        }

        private uint ReadDeltaTime()
        {
            uint time = ReadVLength();
            Debug.Print("Delta Time: " + time.ToString());
            return time;
        }

        private uint ReadVLength()
        {
            uint length = 0;
            byte data;

            for (int i = 0; i < 4; i++)
            {
                data = GetNextByte();
                length += (uint)(data & 0x7F);
                if ((data & 0x80) == 0)
                {
                    break;
                }
                else
                {
                    length <<= 7;
                }
            }

            return length;
        }

        private byte GetNextByte()
        {
            if (index < this.fileBuffer.Length)
            {
                //Debug.Print("Byte: " + this.fileBuffer[index].ToString());
                return this.fileBuffer[index++];
            }
            else
            {
                return 0x00;
            }
        }
    }
}
